/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Fabian Peller-Konrad (fabian dot peller-konrad at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

// Simox
#include <SimoxUtility/meta/type_name.h>

// Header
#include "Resolver.h"

namespace armarx::aron
{
    type::Descriptor Resolver::GetDescriptor(const type::AronTypePtr& a)
    {
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    { \
        type::Aron##upperType##Ptr aron = type::Aron##upperType##Ptr::dynamicCast(a); \
        if(aron) \
        { \
            return type::Descriptor::e##upperType; \
        } \
    }
        HANDLE_ALL_ARON_TYPES
#undef RUN_ARON_MACRO
        throw error::AronNotValidException("Resolver", "GetTypeForAronAbstractType", "Could not cast AronType to a valid type", a);
    }

    data::Descriptor Resolver::GetDescriptor(const data::AronDataPtr& a)
    {
#define RUN_ARON_MACRO(upperType, lowerType, capsType) \
    { \
        data::Aron##upperType##Ptr aron = data::Aron##upperType##Ptr::dynamicCast(a); \
        if(aron) \
        { \
            return data::Descriptor::e##upperType; \
        } \
    }
        HANDLE_ALL_ARON_DATA
#undef RUN_ARON_MACRO
        throw error::AronNotValidException("Resolver", "GetTypeForAronData", "Could not cast AronData to a valid type", a);
    }

    bool Resolver::Correspond(const type::Descriptor& t, const data::Descriptor& d)
    {
#define RUN_ARON_MACRO(typeUpperType, typeLowerType, typeCapsType, dataUpperType, dataLowerType, dataCapsType) \
    if(t == type::Descriptor::e##typeUpperType && d == data::Descriptor::e##dataUpperType) \
    { \
        return true; \
    }

        HANDLE_ALL_CORRESPONDING
#undef RUN_ARON_MACRO
        return false;
    }

    type::Descriptor Resolver::GetFirstIfCorrespond(const type::Descriptor& t, const data::Descriptor& d)
    {
        if (Correspond(t, d))
        {
            return t;
        }
        return GetCorresponding(d);
    }

    data::Descriptor Resolver::GetCorresponding(const type::Descriptor& d)
    {
#define RUN_ARON_MACRO(typeUpperType, typeLowerType, typeCapsType, dataUpperType, dataLowerType, dataCapsType) \
    { \
        if(d == type::Descriptor::e##typeUpperType) \
        { \
            return data::Descriptor::e##dataUpperType; \
        } \
    }
        HANDLE_ALL_CORRESPONDING
#undef RUN_ARON_MACRO
        return data::Descriptor::eUnknown;
    }

    type::Descriptor Resolver::GetCorresponding(const data::Descriptor& d)
    {
#define RUN_ARON_MACRO(typeUpperType, typeLowerType, typeCapsType, dataUpperType, dataLowerType, dataCapsType) \
    { \
        if(d == data::Descriptor::e##dataUpperType) \
        { \
            return type::Descriptor::e##typeUpperType; \
        } \
    }
        HANDLE_ALL_CORRESPONDING
#undef RUN_ARON_MACRO
        return type::Descriptor::eUnknown;
    }

    bool Resolver::AreRelated(const data::Descriptor& d, const data::Descriptor& d2)
    {
        if (d == d2)
        {
            return true;
        }
#define RUN_ARON_MACRO(data1UpperType, data1LowerType, data1CapsType, data2UpperType, data2LowerType, data2CapsType) \
    if( \
        (d == data::Descriptor::e##data1UpperType && d2 == data::Descriptor::e##data2UpperType) || \
        (d2 == data::Descriptor::e##data1UpperType && d == data::Descriptor::e##data2UpperType) \
      ) \
    { \
        return true; \
    }

        HANDLE_ARON_PRIMITIVE_DATA_RELATED
#undef RUN_ARON_MACRO
        return false;
    }

    bool Resolver::AreRelated(const type::Descriptor& t, const type::Descriptor& t2)
    {
        if (t == t2)
        {
            return true;
        }
#define RUN_ARON_MACRO(type1UpperType, type1LowerType, type1CapsType, type2UpperType, type2LowerType, type2CapsType) \
    if( \
        (t == type::Descriptor::e##type1UpperType && t2 == type::Descriptor::e##type2UpperType) || \
        (t2 == type::Descriptor::e##type1UpperType && t == type::Descriptor::e##type2UpperType) \
      ) \
    { \
        return true; \
    }

        HANDLE_ARON_PRIMITIVE_DATA_RELATED
#undef RUN_ARON_MACRO
        return false;
    }

    data::Descriptor Resolver::GetFirstIfRelated(const data::Descriptor& d, const data::Descriptor& d2)
    {
        if (AreRelated(d, d2))
        {
            return d;
        }
        return d2;
    }

    type::Descriptor Resolver::GetFirstIfRelated(const type::Descriptor& t, const type::Descriptor& t2)
    {
        if (AreRelated(t, t2))
        {
            return t;
        }
        return t2;
    }
}
